/* File obslog$Src/add_fast.c
 * January 16, 2007
 * By Tod Lauer, modified by Doug Mink
 *
 *	Format IRAF hselect output for FAST spectra data.
 *	This routine parses the hselect line, and makes a list of
 *	strings that can be picked up for the master log program

 *	Author:	Tod R. Lauer	4/9/94
 *	Modified to read from image file by Doug Mink, 2001-05-18
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "fitsfile.h"
#include "wcstab.h"

#define MAXC 640		/* Length of input line.		*/
#define MAXW 640		/* Maximum word length.			*/

int
add_fast (imagefile, verbose)

char	*imagefile;
int	verbose;
{
	int	hr, mn, sec, ncross, nbin;
	int	lw, idisp;
	float 	rsec;
	double	ra, dec;
	char	word[MAXW],
		texw[MAXW],
		texw2[MAXW],
		texw3[MAXW],
		line[MAXC],
		*endword,
		*header,
		*GetFITShead();
	int debug = 0;

	/* Read the image file header */
	if ((header = GetFITShead (imagefile)) == NULL) {
            fprintf (stderr, "ADD_FAST: Cannot read image header %s\n",
		     imagefile);
            return (-1);
            }
	else if (debug)
	    fprintf (stderr, "ADD_FAST: image file %s\n", imagefile);
 
	/* Delete the extension from the filename and write to start of line */
	strcpy (word, imagefile);
	lw = strlen (word);
	endword = word + lw - 1;
	while (*endword != '.')
	    *endword-- = (char) 0;
	*endword = (char) 0;
	sprintf (texw,"%.32s", word);
	texcheck(texw);
	strcpy (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: image %s\n",texw);

	/* Exposure time to nearest second */
	hgetr4 (header, "EXPTIME", &rsec);
 	sprintf (texw," %.0f", rsec);
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: exposure time = %s\n",texw);

	/* Round the UT start time to the nearest minute */
	hgets (header, "UT", MAXW, word);
	sscanf (word,"%2d:%2d:%2d",&hr,&mn,&sec);
	if (sec >= 30)
	    mn++;
	if (mn >= 60) {
	    mn -= 60;
	    hr++;
	    }
	if (hr >= 24)
	    hr -= 24;
	sprintf (texw," %02d:%02d", hr, mn);
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: UT time = %s\n",texw);

	/* Right ascension */
	if (!hgets (header, "RRA", MAXW, word))
	    hgets (header, "RA", MAXW, word);
	ra = str2ra (word);
	ra2str (texw2, MAXW, ra, 1);
	sprintf (texw," %s", texw2); 
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: RA = %s\n",texw);

	/* Declination */
	if (!hgets (header, "RDEC", MAXW, word))
	    hgets (header, "DEC", MAXW, word);
	dec = str2dec (word);
	dec2str (texw3, MAXW, dec, 0);
        sprintf (texw," %s", texw3); 
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: Dec = %s\n",texw);

	/* Grating */
	strcpy (word, "0");
	!hgets (header, "DISPERSE", MAXW, word);
	sprintf (texw," %s", word);
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: Grating = %s\n",texw);

	/* Rotator angle */
	strcpy (word, "0");
	rsec = 0.0;
	if (!hgetr4 (header, "ROTANGLE", &rsec))
	    hgetr4 (header, "POSANGLE", &rsec);
 	sprintf (texw," %.0f ", rsec);
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: Angle = %s\n",texw);

	/* Slit size */
	strcpy (word, "0");
	hgetr4 (header, "APERTURE", &rsec);
 	sprintf (texw," %.1f ", rsec);
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: Aperture = %s\n",texw);

	/* Program */
	strcpy (word, "none");
	hgets (header, "PROGRAM", MAXW, word);
	sprintf (texw," \"%.14s\"", word);
	texcheck(texw);   
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: program = %s\n",texw);

	/* Binning */
	hgeti4 (header, "DISPAXIS", &idisp);
	if (idisp == 1)
	    hgeti4 (header, "NAXIS2", &ncross);
	else
	    hgeti4 (header, "NAXIS1", &ncross);
	if (ncross < 100)
	    nbin = 4;
	else if (ncross < 150)
	    nbin = 3;
	else if (ncross < 200)
	    nbin = 2;
	else
	    nbin = 1;
 	sprintf (texw," %d", nbin);
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: binning = %s\n",texw);

	/* Read observation comments from REMARK keyword */
	strcpy (word, " ");
	hgets (header, "REMARK", MAXW, word);
	sprintf (texw," \"%s\"", word);
	texcheck(texw);   
	strcat (line, " ");
	strcat (line, texw);
	if (debug)
	    fprintf (stderr, "ADD_FAST: comment = %s\n",texw);

	/* Print completed line to standard output */
	printf ("%s\n", line); 
	if (verbose)
	    fprintf (stderr, "%s\n", line);
	return (0);
}
/* May 18 2001	Change so only one line is printed in verbose mode
 *
 * Nov 13 2002	Change keyword from POSANGLE to ROTANGLE
 *
 * Jan 12 2007	Add include statements
 * Jan 16 2007	Convert RA and Dec to double then back to string
 * Jan 16 2007	Print output line to stderr if verbose mode
 */
