/*** File libwcs/tabread.c
 *** January 12, 2007
 *** By Doug Mink, dmink@cfa.harvard.edu
 *** Harvard-Smithsonian Center for Astrophysics
 *** Copyright (C) 1996-2007
 *** Smithsonian Astrophysical Observatory, Cambridge, MA, USA

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.
    
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Correspondence concerning WCSTools should be addressed as follows:
           Internet email: dmink@cfa.harvard.edu
           Postal address: Doug Mink
                           Smithsonian Astrophysical Observatory
                           60 Garden St.
                           Cambridge, MA 02138 USA
 */

/* struct TabTable *tabopen()	Open tab table, returning number of entries
 * char *tabline()	Get tab table entry for one line
 * double tabgetra()	Return double right ascension in degrees
 * double tabgetdec()	Return double declination in degrees
 * double tabgetr8()	Return 8-byte floating point number from tab table line
 * int tabgeti4()	Return 4-byte integer from tab table line
 * int tabgetk()	Return character entry from tab table line for column
 * int tabgetc()	Return n'th character entry from tab table line
 * int tabparse()	Make a table of column headings
 * int tabcol()		Search a table of column headings for a particlar entry
 * int tabsize()	Return length of file in bytes
 * int istab()		Return 1 if first line of file contains a tab, else 0
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include "fitsfile.h"
#include "wcstab.h"

static char newline = 10;
static char tab = 9;

#define ABS(a) ((a) < 0 ? (-(a)) : (a))

static int tabsize();
static int nndec = 0;
static char *taberr;
char *gettaberr ()
{ return (taberr); }

int gettabndec()
{ return (nndec); }

/* TABOPEN -- Open tab table file, returning number of entries */

struct TabTable *
tabopen (tabfile, nbbuff)

char	*tabfile;	/* Tab table catalog file name */
int	nbbuff;		/* Number of bytes in buffer; 0=read whole file */
{
    FILE *fcat;
    int nr, lfile, lname, lline;
    char *tabnew, *tabline, *lastline;
    char *tabcomma, *nextline;
    char *thisname, *tabname;
    int thistab, itab, nchar, nbtab;
    int formfeed = (char) 12;
    struct TabTable *tabtable;

    tabcomma = NULL;
    if (taberr != NULL) {
	free (taberr);
	taberr = NULL;
	}

    tabname = NULL;
    if (!strcmp (tabfile, "stdin")) {
	lfile = 100000;
	fcat = stdin;
	}
    else {

	/* Separate table name from file name, if necessary */
	if ((tabcomma = strchr (tabfile, ',')) != NULL) {
	    tabname = (char *) calloc (1,64);
	    strcpy (tabname, tabcomma+1);
	    *tabcomma = (char) 0;
	    }

	/* Find length of tab table catalog */
	lfile = tabsize (tabfile);
	if (nbbuff > 1 && nbbuff < lfile)
	    lfile = nbbuff;
	else if (nbbuff == 1 && lfile > 10000)
	    lfile = 10000;
	if (lfile < 1) {
	    taberr = (char *) calloc (64 + strlen (tabfile), 1);
	    sprintf (taberr,"TABOPEN: Tab table file %s has no entries",
		     tabfile);
	    if (tabcomma != NULL) *tabcomma = ',';
	    return (NULL);
	    }

	/* Open tab table catalog */
	if (!(fcat = fopen (tabfile, "r"))) {
	    taberr = (char *) calloc (64 + strlen (tabfile), 1);
	    sprintf (taberr,"TABOPEN: Tab table file %s cannot be read",
		     tabfile);
	    if (tabcomma != NULL) *tabcomma = ',';
	    return (NULL);
	    }
	}

    /* Allocate tab table structure */
    nbtab = sizeof(struct TabTable);
    if ((tabtable=(struct TabTable *) calloc(1,nbtab)) == NULL){
	taberr = (char *) calloc (64 + strlen (tabfile), 1);
	sprintf (taberr,"TABOPEN: cannot allocate Tab Table structure for %s",
		 tabfile);
	if (tabcomma != NULL) *tabcomma = ',';
	return (NULL);
	}

    tabtable->tabname = tabname;

    /* Allocate space in structure for filename and save it */
    lname = strlen (tabfile) + 2;
    if ((tabtable->filename = (char *)calloc (1, lname)) == NULL) {
	taberr = (char *) calloc (64 + strlen (tabfile), 1);
	sprintf (taberr,"TABOPEN: cannot allocate filename %s in structure",
		 tabfile);
	(void) fclose (fcat);
	tabclose (tabtable);
	if (tabcomma != NULL) *tabcomma = ',';
	return (NULL);
	}
    strcpy (tabtable->filename, tabfile);

    /* Allocate buffer to hold entire catalog (or buffer length) and read it */
    if ((tabtable->tabbuff = (char *) calloc (1, lfile+2)) == NULL) {
	taberr = (char *) calloc (64 + strlen (tabfile), 1);
	sprintf (taberr,"TABOPEN: cannot allocate buffer for tab table %s",
		 tabfile);
	(void) fclose (fcat);
	tabclose (tabtable);
	if (tabcomma != NULL) *tabcomma = ',';
	return (NULL);
	}
    else {
	nr = fread (tabtable->tabbuff, 1, lfile, fcat);
	if (fcat != stdin && nr < lfile) {
	    fprintf (stderr,"TABOPEN: read only %d / %d bytes of file %s\n",
		     nr, lfile, tabfile);
	    (void) fclose (fcat);
	    tabclose (tabtable);
	    if (tabcomma != NULL) *tabcomma = ',';
	    return (NULL);
	    }
	tabtable->tabbuff[lfile] = (char) 0;

	/* Check for named table within a file */
	if (tabname != NULL) {
	    if (isnum (tabname)) {
		itab = atoi (tabname);
		thisname = tabtable->tabbuff;
		thistab = 1;
		if (itab > 1) {
		    while (thistab < itab && thisname != NULL) {
			thisname = strchr (thisname, formfeed);
			if (thisname != NULL)
			    thisname++;
			thistab++;
			}
		    }
		if (thisname == NULL) {
		    fprintf (stderr, "GETTAB:  There are < %d tables in %s\n",
			itab, tabfile);
		    return (NULL);
		    }
		while (*thisname==' ' || *thisname==newline ||
		       *thisname==formfeed || *thisname==(char)13)
		    thisname++;
		tabline = strchr (thisname, newline);
		if (tabline != NULL) {
		    nchar = tabline - thisname;
		    if (strchr (thisname, tab) > tabline)
			strncpy (tabtable->tabname, thisname, nchar);
		    }
		}
	    else {
		lname = strlen (tabname);
		thisname = tabtable->tabbuff;
		while (*thisname != (char) 0) {
		    while (*thisname==' ' || *thisname==newline ||
		   	   *thisname==formfeed || *thisname==(char)13)
			thisname++;
		    if (!strncmp (tabname, thisname, lname))
			break;
		    else
			thisname = strchr (thisname, formfeed);
		    }
		}
	    if (thisname == NULL) {
		fprintf (stderr, "TABOPEN: table %s in file %s not found\n",
			 tabname, tabfile);
		if (tabcomma != NULL) *tabcomma = ',';
		return (NULL);
		}
	    else
		tabtable->tabheader = strchr (thisname, newline) + 1;
	    }
	else
	    tabtable->tabheader = tabtable->tabbuff;

	tabline = tabtable->tabheader;
	lastline = NULL;
	while (*tabline!='-' && tabline!=NULL && tabline < tabtable->tabbuff+lfile) {
	    lastline = tabline;
	    if ((tabline = strchr (lastline,newline)) != NULL)
		tabline++;
	    else
		break;
	    }
	if (tabline == NULL || *tabline != '-') {
	    taberr = (char *) calloc (64 + strlen (tabfile), 1);
	    sprintf (taberr,"TABOPEN: No - line in tab table %s",tabfile);
	    (void) fclose (fcat);
	    tabclose (tabtable);
	    if (tabcomma != NULL) *tabcomma = ',';
	    return (NULL);
	    }
	tabtable->tabhead = lastline;
	tabtable->tabdash = tabline;
	tabtable->tabdata = strchr (tabline, newline) + 1;

	/* Extract positions of keywords we will want to use */
	if (!tabparse (tabtable)) {
	    fprintf (stderr,"TABOPEN: No columns in tab table %s\n",tabfile);
	    (void) fclose (fcat);
	    tabclose (tabtable);
	    if (tabcomma != NULL) *tabcomma = ',';
	    return (NULL);
	    }

    /* Enumerate entries in tab table catalog by counting newlines */
	if (nbbuff > 0)
	    tabtable->nlines = 10000000;
	else {
	    tabnew = tabtable->tabdata;
	    tabtable->nlines = 0;
	    while ((tabnew = strchr (tabnew, newline)) != NULL) {
		tabnew = tabnew + 1;
		tabtable->nlines = tabtable->nlines + 1;
		if (*tabnew == formfeed)
		    break;
		}
	    }
	}

    /* Set up line buffer, and put first line in it */
    if (nbbuff > 0) {
	tabtable->lhead = tabtable->tabdata - tabtable->tabheader;
	tabtable->tcat = fcat;
	nextline = strchr (tabtable->tabdata, newline) + 1;
	tabtable->lline = (nextline - tabtable->tabdata) * 2;
	tabtable->tabline = (char *) calloc (tabtable->lline, 1);
	fseek (tabtable->tcat, (long) tabtable->lhead, SEEK_SET);
	fgets (tabtable->tabline, tabtable->lline, tabtable->tcat);
	lline = strlen (tabtable->tabline);
	if (tabtable->tabline[lline-1] < 32)
	     tabtable->tabline[lline-1] = (char) 0;
	tabtable->tabdata = tabtable->tabline;
	}

    /* Close catalog file if not reading one line at a time */
    else {
	tabtable->lhead = 0;
	(void) fclose (fcat);
	tabtable->tcat = NULL;
	}

    tabtable->tabline = tabtable->tabdata;
    tabtable->iline = 1;
    if (tabcomma != NULL) *tabcomma = ',';
    return (tabtable);
}


void
tabclose (tabtable)

    struct TabTable *tabtable;
{
    if (tabtable != NULL) {
	if (tabtable->filename != NULL) free (tabtable->filename);
	if (tabtable->tabname != NULL) free (tabtable->tabname);
	if (tabtable->tabbuff != NULL) free (tabtable->tabbuff);
	if (tabtable->colname != NULL) free (tabtable->colname);
	if (tabtable->lcol != NULL) free (tabtable->lcol);
	if (tabtable->lcfld != NULL) free (tabtable->lcfld);
	if (tabtable->tcat != NULL) fclose (tabtable->tcat);
	free (tabtable);
	}
    return;
}


/* TABLINE -- Get tab table entry for one line;
	      return NULL if unsuccessful */

char *
gettabline (tabtable, iline)

struct TabTable *tabtable;	/* Tab table structure */
int iline;	/* Line sequence number in tab table */
{
    char *nextline = tabtable->tabline;
    char *next;
    int lline, i;

    /* Return NULL if tab table has not been opened */
    if (tabtable == NULL)
	return (NULL);

    /* Read one line at a time if file is still open */
    if (tabtable->tcat != NULL) {

	/* Return current line from buffer */
	if (iline == tabtable->iline)
	    return (tabtable->tabline);

	/* Read next line from file */
	if (iline < 1 || iline > tabtable->iline) {
	    for (i = tabtable->iline; i < iline; i++) {
		next = fgets (tabtable->tabline, tabtable->lline, tabtable->tcat);
		if (next == NULL || *next == EOF)
		    return (NULL);
		tabtable->iline++;
		}
	    lline = strlen (tabtable->tabline);
	    if (lline < 2)
		return (NULL);
	    if (tabtable->tabline[lline-1] < 32)
		tabtable->tabline[lline-1] = (char) 0;
	    }
	else if (iline < tabtable->iline) {
	    fseek (tabtable->tcat, (long) tabtable->lhead, SEEK_SET);
	    tabtable->iline = 0;
	    for (i = tabtable->iline; i < iline; i++) {
		fgets (tabtable->tabline, tabtable->lline, tabtable->tcat);
		tabtable->iline++;
		}
	    lline = strlen (tabtable->tabline);
	    if (tabtable->tabline[lline-1] < 32)
		tabtable->tabline[lline-1] = (char) 0;
	    }
	return (tabtable->tabline);
	}

    /* Return NULL if trying to read past last line */
    if (iline > tabtable->nlines) {
	fprintf (stderr, "TABLINE:  line %d is not in table\n",iline);
	return (NULL);
	}

    /* If iline is 0 or less, just read next line from table */
    else if (iline < 1 && nextline) {
	tabtable->iline++;
	if (tabtable->iline > tabtable->nlines) {
	    fprintf (stderr, "TABLINE:  line %d is not in table\n",iline);
	    return (NULL);
	    }
	nextline = strchr (nextline, newline) + 1;
	}

    /* If iline is before current line, read from start of file */
    else if (iline < tabtable->iline) {
	tabtable->iline = 1;
	tabtable->tabline = tabtable->tabdata;
	while (tabtable->iline < iline) {
	    tabtable->tabline = strchr (tabtable->tabline, newline) + 1;
	    tabtable->iline ++;
	    }
	}
    /* If iline is after current line, read forward */
    else if (iline > tabtable->iline) {
	while (tabtable->iline < iline) {
	    tabtable->tabline = strchr (tabtable->tabline, newline) + 1;
	    tabtable->iline ++;
	    }
	}

    return (tabtable->tabline);
}


/* TABGETRA -- returns double right ascension in degrees */

double
tabgetra (tabtok, ientry)

struct Tokens *tabtok;	/* Line token structure */
int	ientry;	/* sequence of entry on line */
{
    char str[24];

    strcpy (str, "0.0");
    if (tabgetc (tabtok, ientry, str, 24))
	return (0.0);
    else
	return (str2ra (str));
}


/* TABGETDEC -- returns double declination in degrees */

double
tabgetdec (tabtok, ientry)

struct Tokens *tabtok;	/* Line token structure */
int	ientry;		/* sequence of entry on line */
{
    char str[24];

    strcpy (str, "0.0");
    if (tabgetc (tabtok, ientry, str, 24))
	return (0.0);
    else
	return (str2dec (str));
}


/* TABGETR8 -- returns 8-byte floating point number from tab table line */

double
tabgetr8 (tabtok, ientry)

struct Tokens *tabtok;	/* Line token structure */
int	ientry;		/* sequence of entry on line */
{
    char str[24];

    strcpy (str, "0.0");
    if (tabgetc (tabtok, ientry, str, 24))
	return (0.0);
    else if (isnum (str))
        return (atof (str));
    else
	return (0.0);
}


/* TABGETI4 -- returns a 4-byte integer from tab table line */

int
tabgeti4 (tabtok, ientry)

struct Tokens *tabtok;	/* Line token structure */
int	ientry;		/* sequence of entry on line */
{
    char str[24];

    strcpy (str, "0");
    if (tabgetc (tabtok, ientry, str, 24))
	return (0);
    else if (isnum (str))
        return ((int) atof (str));
    else
	return (0);
}


/* TABGETK -- returns a character entry from tab table line for named column */

int
tabgetk (tabtable, tabtok, keyword, string, maxchar)

struct TabTable *tabtable;	/* Tab table structure */
struct Tokens *tabtok;		/* Line token structure */
char	*keyword;		/* column header of desired value */
char	*string;		/* character string (returned) */
int	maxchar;	/* Maximum number of characters in returned string */
{
    int ientry = tabcol (tabtable, keyword);

    return (tabgetc (tabtok, ientry, string, maxchar));
}


/* TABGETC -- returns n'th entry from tab table line as character string */

int
tabgetc (tabtok, ientry, string, maxchar)

struct Tokens *tabtok;	/* Line token structure */
int	ientry;		/* Sequence of entry on line (1-ncol) */
char	*string;	/* Character string (returned) */
int	maxchar;	/* Maximum number of characters in returned string */
{

    if (ientry > tabtok->ntok)
	return (0);
    else if (getoken (tabtok, ientry, string, maxchar))
	return (0);
    else
	return (-1);
}


/* TABPARSE -- Make a table of column headings */

int
tabparse (tabtable)

struct TabTable *tabtable;	/* Tab table structure */
{
    char *colhead;	/* Column heading first character */
    char *endcol;	/* Column heading last character */
    char *headlast;
    char *hyphens;
    char *hyphlast;
    char *nextab;
    int icol;
    int nbytes, nba;

    /* Return if no column names in header */
    headlast = strchr (tabtable->tabhead, newline);
    if (headlast == tabtable->tabhead)
	return (0);

    /* Count columns in table header */
    tabtable->ncols = 1;
    for (colhead = tabtable->tabhead; colhead < headlast; colhead++) {
	if (*colhead == tab)
	    tabtable->ncols++;
	}

    /* Tabulate column names */
    nbytes = tabtable->ncols * sizeof (char *);
    nba = nbytes / 64;
    if (nbytes > nba * 64)
	nba = (nba + 1) * 64;
    else
	nba = nba * 64;
    tabtable->colname = (char **)calloc (tabtable->ncols, sizeof (char *));
    tabtable->lcol = (int *) calloc (tabtable->ncols, sizeof (int));
    colhead = tabtable->tabhead;
    for (icol = 0; icol < tabtable->ncols; icol++) {
	nextab = strchr (colhead, tab);
	if (nextab < headlast)
	    endcol = nextab - 1;
	else
	    endcol = headlast - 1;
	while (*endcol == ' ')
	    endcol = endcol - 1;
	tabtable->lcol[icol] = (int) (endcol - colhead) + 1;
	tabtable->colname[icol] = colhead;
	colhead = nextab + 1;
	if (colhead > headlast)
	    break;
	}

    /* Tabulate field widths */
    hyphens = headlast + 1;
    hyphlast = strchr (hyphens, newline);
    if (hyphlast == hyphens)
	return (0);
    tabtable->lcfld = (int *) calloc (tabtable->ncols, sizeof (int));
    colhead = hyphens;
    for (icol = 0; icol < tabtable->ncols; icol++) {
	if ((nextab = strchr (colhead, tab)) == NULL)
	    endcol = hyphlast - 1;
	else
	    endcol = nextab - 1;
	tabtable->lcfld[icol] = (int) (endcol - colhead) + 1;
	if (nextab != NULL)
	    colhead = nextab + 1;
	else
	    break;
	}

    return (tabtable->ncols);
}


/* Search table of column headings for a particlar entry (case-dependent) */

int
tabcol (tabtable, keyword)

struct TabTable *tabtable;	/* Tab table structure */
char	*keyword;		/* Column heading to find */

{
    int i, lkey, lcol;
    lkey = strlen (keyword);

    for (i = 0; i < tabtable->ncols; i++) {
	lcol = tabtable->lcol[i];
	if (lcol == lkey &&
	    !strncasecmp (keyword, tabtable->colname[i], lcol)) {
	    return (i + 1);
	    }
	}
    return (0);
}


/* TABSIZE -- return size of file in bytes */

static int
tabsize (filename)

char	*filename;	/* Name of file for which to find size */
{
    FILE *diskfile;
    long filesize;

    /* Open file */
    if ((diskfile = fopen (filename, "r")) == NULL)
	return (-1);

    /* Move to end of the file */
    if (fseek (diskfile, 0, 2) == 0)

 	/* Position is the size of the file */
	filesize = ftell (diskfile);

    else
	filesize = -1;

    fclose (diskfile);

    return (filesize);
}


/* ISTAB -- Return 1 if tab table file, else 0 */

int
istab (filename)

char    *filename;      /* Name of file to check */
{
    struct TabTable *tabtable;

    /* First check file extension */
    if (strsrch (filename, ".tab"))
	return (1);

    /* If no .tab file extension, try opening the file */
    else {
	if ((tabtable = tabopen (filename, 1000)) != NULL) {
	    tabclose (tabtable);
	    return (1);
	    }
	else
	    return (0);
	}
}

/* Jul 18 1996	New subroutines
 * Aug  6 1996	Remove unused variables after lint
 * Aug  8 1996	Fix bugs in entry reading and logging
 * Oct 15 1996  Add comparison when testing an assignment
 * Nov  5 1996	Drop unnecessary static declarations
 * Nov 13 1996	Return no more than maximum star number
 * Nov 13 1996	Write all error messages to stderr with subroutine names
 * Nov 15 1996  Implement search radius; change input arguments
 * Nov 19 1996	Allow lower case column headings
 * Dec 18 1996	Add UJCRNUM to read specified catalog entries
 * Dec 18 1996  Keep closest stars, not brightest, if searching within radius
 *
 * Mar 20 1997	Clean up code in TABRNUM
 * May  7 1997	Set entry number to zero if column not found
 * May 29 1997	Add TABPARSE and TABCOL to more easily extract specific columns
 * May 29 1997	Add TABCLOSE to free memory from outside this file
 * Jun  4 1997	Set ID to sequence number in table if no ID/id entry present
 *
 * Jun  2 1998	Fix bug parsing last column of header
 * Jun 24 1998	Add string lengths to ra2str() and dec2str() calls
 * Sep 16 1998	Use limiting radius correctly
 * Sep 22 1998	Convert to output coordinate system
 * Oct 15 1998	Add tabsize() and istab()
 * Oct 21 1998	Add tabrkey() to read values of keyword for list of stars
 * Oct 29 1998	Correctly assign numbers when too many stars are found
 * Oct 30 1998	Fix istab() to check only first line of file
 * Dec  8 1998	Do not declare tabsize() static

 * Jan 20 1999	Add tabcatopen() and keep table info in structure, not global
 * Jan 25 1999	Add lcfld to structure to keep track of field widths
 * Jan 29 1999	Add current line number and pointer to table structure
 * Feb  2 1999	Allow for equinox other than 2000.0 in tab table header
 * Feb  2 1999	Add tabhgetc() to read char string values from tab table header
 * Feb 17 1999	Increase maximum line in istab() from 80 to 1024
 * Mar  2 1999	Fix bugs calling tabhgetx()
 * Mar  2 1999	Rewrite tabhgetx() to use tabhgetc() for all header reading
 * May 28 1999	Add tabcatopen() and tabstar() and use them
 * Jun  3 1999	Fix bug so header parameters are read correctly
 * Jun 16 1999	Use SearchLim()
 * Aug 16 1999	Fix bug to fix failure to search across 0:00 RA
 * Aug 25 1999  Return real number of stars from tabread()
 * Sep 10 1999	Fix bug setting equinox and coordinate system in tabcatopen()
 * Sep 10 1999	Set additional keyword selection with subroutine
 * Sep 13 1999	Fix comment for tabstar()
 * Sep 16 1999	Fix bug which didn't always return closest stars
 * Sep 16 1999	Add distsort argument so brightest stars in circle works, too
 * Oct 21 1999	Clean up code after lint
 * Oct 25 1999	Fix subroutine declaration inconsistency
 * Oct 25 1999	Replace malloc() calls with calloc()
 * Oct 29 1999	Add tabxyread() for image catalogs
 * Nov 23 1999	Improve error checking on Starbase tables; istab() opens file
 * Nov 30 1999	Fix bugs found when compiling under SunOS 4.1.3
 *
 * Jan  4 2000	Always close file and print error message on tabopen() failures
 * Jan  6 2000	If "id" not found, try heading with "_id" to catch scat output
 * Jan 10 2000	Add second magnitude; save column headers in catalog structure
 * Feb 10 2000	Implement proper motion in source catalogs
 * Feb 10 2000	Accept first mag-containing column as first magnitude
 * Feb 10 2000	Clean up id reading: no. decimals, non-numeric id
 * Feb 14 2000	Save table opening errors in string
 * Feb 16 2000	Lengthen short calloc() lengths
 * Feb 16 2000	Pad tabbuff with 2 nulls so end can be found
 * Mar 10 2000	Return proper motions from tabread() and tabrnum()
 * Mar 13 2000	Do not free tabtable structure if it is null
 * Mar 27 2000	Clean up code after lint
 * May 26 2000	Add ability to read named tables in a multi-table file
 * Jun 26 2000	Add coordinate system to SearchLim() arguments
 * Jul  5 2000	Check for additional column heading variations
 * Jul 10 2000	Deal with number of decimal places and name/number in tabcatopen()
 * Jul 12 2000	Add star catalog data structure to tabread() argument list
 * Jul 13 2000	Use nndec header parameter to optionally set decimal places
 * Jul 25 2000	Pass star catalog address of data structure address
 * Aug  3 2000	Skip first character of ID if rest is number
 * Aug  3 2000	If no decimal point in numeric ID, set ndec to zero
 * Sep 27 2000	Use first column with name if no id column is found
 * Oct 26 2000	Add proper motion in seconds and arcseconds per century
 * Oct 31 2000	Add proper motion in milliseconds of time per year
 * Nov 22 2000	Add tabtable argument to tabcatopen() for URL search returns
 * Nov 28 2000	Add starcat as argument to tabrnum() as well as tabread()
 * Nov 29 2000	Do not set tmagb if it is null; set type if present
 * Nov 30 2000	Add spectral type as possible column
 * Dec  1 2000	Add field as synonym for plate
 * Dec 18 2000	Clean up after lint
 * Dec 29 2000	Clean up after lint
 *
 * May 29 2001	Save length of identifier in catalog structure
 * May 30 2001	Rewrite to deal with up to 3 magnitudes
 * Jun 11 2001	Add one-line-at-a-time catalog reading from tabcatopen()
 * Jun 14 2001	In tabopen(), use actual file size if less than buffer size
 * Jun 14 2001	In tabcol() make sure lengths are equal, too
 * Jun 18 2001	Fix bug finding of length of token string extracted
 * Jun 19 2001	Use setoken() to parse each catalog line
 * Jun 20 2001	Add fourth magnitude for GSC II
 * Jun 25 2001	Check fgets() return in gettabline()
 * Jun 25 2001	Print star read errors only in verbose mode
 * Jun 28 2001	If no proper motion unit is set, set proper motion to 0.0
 * Jun 28 2001	If first magnitude is 99.90, sort by second magnitude
 * Jun 28 2001	Up default line limit from 1 million to 10 million
 * Jul  2 2001	Fix order of 3rd and 4th magnitudes
 * Jul 20 2001	Return magnitude as well as flux from tabxyread()
 * Aug  8 2001	Return radial velocity as additional magnitude
 * Aug 17 2001	Fix bug reading radial velocity
 * Aug 21 2001	Check numbers using isnum() in tabgetr8() and tabgeti4()
 * Aug 21 2001	Add starcat to tabrkey() argument list
 * Aug 21 2001	Read object name into tkey if it is present
 * Sep 11 2001	Allow an arbitrary number of magnitudes
 * Sep 11 2001  Add sort magnitude argument to tabread()
 * Sep 19 2001	Drop fitshead.h; it is in wcs.h
 * Sep 20 2001	Clean up pointers for Alpha and Linux; drop tabgetpm()
 * Sep 20 2001	Change _ to . in ESO server USNO-A2.0 numbers
 * Sep 21 2001	Rearrange ESO server GSC numbers
 * Oct 12 2001	Check for additional column names for magnitude
 * Oct 15 2001	Read first star only once, compute relative flux in tabxyread() (bug fix)
 * Oct 16 2001	Add pointer to line of dashes to table structure
 * Dec  3 2001	Initialize keyword search variables in tabhgetc()
 *
 * May  6 2002	Allow object names to be up to 79 characters long
 * Aug  5 2002	Deal correctly with magnitude-less catalogs
 * Aug  5 2002	Add magu, magb, magv for UBV magnitudes
 * Aug  6 2002	Pass through magnitude keywords
 * Aug  6 2002	Return initial string if token not found by tabgetc()
 * Oct 30 2002	Add code to pass epoch as a final magnitude (but not RV, yet)
 *
 * Jan 26 2003	Add code to pass USNO-B1.0 pm quality and no. of ids
 * Jan 28 2003	Improve spatial position test
 * Feb  4 2003	Compare character to 0, not NULL
 * Mar 11 2003	Fix limit setting
 * Apr  3 2003	Drop unused variables after lint
 * May 28 2003	Read long and lat if radecsys is "galactic" or "ecliptic"
 * Jun  2 003	Divide by cos(dec) for arcsec and mas RA proper motions
 * Aug 22 2003	Add radi argument for inner edge of search annulus
 * Sep 25 2003	Add tabbin() to fill an image with sources
 * Oct  6 2003	Update tabread() and tabbin() for improved RefLim()
 * Nov 18 2003	Initialize image size and bits/pixel from header in tabbin()
 * Nov 22 2003	Add GSC II object class (c) as possible content for tpeak
 * Dec  4 2003	Add default of arcsec/century as proper motion unit
 * Dec 10 2003	If magnitude ends in L and is a number, add 100.0
 * Dec 12 2003	Fix bug in wcs2pix() call in tabbin()
 *
 * Jan  5 2003	If more than 15 digits in numberic ID, drop excess off front
 *
 * Jan 12 2007	Drop unused subroutines tabhget*(), tabcont()
 */
